# -*- coding: utf-8 -*-

"""Setup for abnf.  Version information is generated by setuptools_scm."""

from setuptools import setup

setup(
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
)
